/*
 *  **************************************************************************
 *
 *  (c) 2012 Nicolas Bonnel
 * 
 *  **************************************************************************
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,  but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details. 
 *  
 *  You should have received a copy of the GNU Lesser General Public License 
 *  along with this program; if not, write to the Free Software Foundation, 
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  **************************************************************************
 * 
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <vector>
#include <sstream>
#include "rmsd.h"
#include <cstring>

static union
{
	double d;
	struct
	{
#ifdef LITTLE_ENDIAN
		int j, i;
#else
		int i, j;
#endif
	} n;
} _eco;

#define EXP_A (1048576/M_LN2)	/* 2^20/LN(2) use 1512775 for integer version */
#define EXP_C 60801	 /* see text for choice of c values */
#define EXP(y) (_eco.n.i = EXP_A*(y) + (1072693248 - EXP_C), _eco.d)

using namespace std;
const float nuVals[] = {0.67f,0.41f,0.24f,0.15f};
const float gapVals[] = {-0.53f,-0.5f};

typedef struct _point{
	float x,y,z;
}point;

typedef struct _seq{
	unsigned int size;
	float * values;
	string residues;
	point *coords;
}seq;

typedef struct _dp{
	unsigned int move;
	float value;
}dp;

string convertInt(int number)
{
   stringstream ss;//create a stringstream
   ss << number;//add number to the stream
   return ss.str();//return a string with the contents of the stream
}


char letterOfResidue(string res){
	if (res=="ALA") return 'A';
	if (res=="ARG") return 'R';
	if (res=="ASN") return 'N';
	if (res=="ASP") return 'D';
	if (res=="CYS") return 'C';
	if (res=="GLN") return 'Q';
	if (res=="GLU") return 'E';
	if (res=="GLY") return 'G';
	if (res=="HIS") return 'H';
	if (res=="ILE") return 'I';
	if (res=="LEU") return 'L';
	if (res=="LYS") return 'K';
	if (res=="MET"|| res=="MSE") return 'M';
	if (res=="PHE") return 'F';
	if (res=="PRO") return 'P';
	if (res=="SER") return 'S';
	if (res=="THR") return 'T';
	if (res=="TRP") return 'W';
	if (res=="TYR") return 'Y';
	if (res=="VAL") return 'V';
	if (res=="ASX") return 'B';
	if (res=="GLX") return 'Z';
	if (res=="XLE") return 'J';
	if (res=="SEC") return 'U';
	if (res=="PYL") return 'O';
	if (res=="UNK") return 'X';
	cout << "error : "<< res << endl;
}

void nwa(unsigned int dimension,seq s1,seq s2,float nu){
	dp ** mat = new dp*[s2.size];
	for (int i=0;i<s2.size;i++)
		mat[i]= new dp[s1.size];
	for (int i=0;i<s2.size;i++)
		mat[i][0].value = 0.0f;
	for (int i=0;i<s1.size;i++)
		mat[0][i].value = 0.0f;

	unsigned int indI, indJ,indIK,indJK;
	float sum;
	float result = 0.0f;
	unsigned int i,j,k,mi,mj;
	float c,m,d;
	for (i=1;i<s1.size;i++){
		indI = dimension*i;
		for (j=1;j<s2.size;j++){
			indJ = dimension*j;
			sum = 0.0f;
			for (k=0;k<dimension;k++){
				indIK = indI+k;
				indJK = indJ+k;
				sum += (abs(s1.values[indIK]-s2.values[indJK])+abs(s1.values[indIK-dimension]-s2.values[indJK-dimension]))+3.0f*abs((s1.values[indIK]-s1.values[indIK-dimension])-(s2.values[indJK]-s2.values[indJK-dimension]));
			}
			c = mat[j-1][i].value;
			d = mat[j][i-1].value;
			m = mat[j-1][i-1].value+(float)EXP(-nu*sum);
			if (c>m){
				if(c>d){
					mat[j][i].value = c;
					mat[j][i].move = 1;
				}else{
					mat[j][i].value = d;
					mat[j][i].move = 3;
				}
			}else{
				if(m>d){
					mat[j][i].value = m;
					mat[j][i].move = 2;
				}else{
					mat[j][i].value = d;
					mat[j][i].move = 3;
				}
			}
		}
	}
	mi = s1.size-1;
	mj = s2.size-1;
	result = mat[s2.size-1][s1.size-1].value/sqrt((float)(s2.size-1)*(s1.size-1));;
	string a1,a2,a3;
	int cptM = 0;
	int aliL = 1;
	vector<point> v1,v2;
	while(mj>0 && mi>0){
		if(mat[mj][mi].move==2){
			a1.insert(0,1,s1.residues[mi]);
			a2.insert(0,1,s2.residues[mj]);
			v1.push_back(s1.coords[mi]);
			v2.push_back(s2.coords[mj]);
			aliL += 1;
			if (s1.residues[mi]==s2.residues[mj]){
				cptM += 1;
				a3.insert(0,1,s2.residues[mj]);
			}else
				a3.insert(0,1,'.');
			mj-=1;
			mi-=1;
		}else if(mat[mj][mi].move==1){
			a1.insert(0,1,'-');
			a2.insert(0,1,s2.residues[mj]);
			a3.insert(0,1,'.');
			mj-=1;
		}else if(mat[mj][mi].move==3){
			a1.insert(0,1,s1.residues[mi]);
			a2.insert(0,1,'-');
			a3.insert(0,1,'.');
			mi-=1;			
		}
	}
	a1.insert(0,1,s1.residues[mi]);
	a2.insert(0,1,s2.residues[mj]);
	v1.push_back(s1.coords[mi]);
	v2.push_back(s2.coords[mj]);
	if (s1.residues[mi]==s2.residues[mj]){
		cptM += 1;
		a3.insert(0,1,s2.residues[mj]);
	}else
		a3.insert(0,1,' ');
	string l1 = convertInt(mi);
	string l2 = convertInt(mj);
	int len = max(l1.length(),l2.length());
	a1.insert(0,1+len-l1.length(),' ');
	a2.insert(0,1+len-l2.length(),' ');
	a3.insert(0,1+len,' ');
	a1.insert(0,convertInt(mi));
	a2.insert(0,convertInt(mj));
	for (int i=0;i<s2.size;i++)
		delete mat[i];
	delete mat;
	cout << a1 << endl << a2 << endl << a3 <<endl;
	cout << "Alignment length: "<< aliL << endl;
	cout << "Percentage of matching residues: "<< cptM*100.0f/aliL << " %"<< endl; 
	cout << "Score: "<< result<< endl;

	double p1[v1.size()][3], p2[v1.size()][3];
	for (int i=0;i<v1.size();i++){
		p1[i][0] = v1[i].x;
		p1[i][1] = v1[i].y;
		p1[i][2] = v1[i].z;
		p2[i][0] = v2[i].x;
		p2[i][1] = v2[i].y;
		p2[i][2] = v2[i].z;
	}
	double rmsd;
	fast_rmsd(p1,p2,v1.size(),&rmsd);
	cout << "RMSD: "<< rmsd<< endl;
}

void swa(unsigned int dimension,seq s1,seq s2,float nu,float gap){
	dp ** mat = new dp*[s2.size];
	for (int i=0;i<s2.size;i++)
		mat[i]= new dp[s1.size];
	for (int i=0;i<s2.size;i++)
		mat[i][0].value = 0.0f;
	for (int i=0;i<s1.size;i++)
		mat[0][i].value = 0.0f;

	unsigned int indI, indJ,indIK,indJK;
	float sum;
	float result = 0.0f;
	unsigned int i,j,k,mi,mj;
	float c,m,d;
	for (i=1;i<s1.size;i++){
		indI = dimension*i;
		for (j=1;j<s2.size;j++){
			indJ = dimension*j;
			sum = 0.0f;
			for (k=0;k<dimension;k++){
				indIK = indI+k;
				indJK = indJ+k;
				sum += (abs(s1.values[indIK]-s2.values[indJK])+abs(s1.values[indIK-dimension]-s2.values[indJK-dimension]))+3.0f*abs((s1.values[indIK]-s1.values[indIK-dimension])-(s2.values[indJK]-s2.values[indJK-dimension]));
			}
			c = mat[j-1][i].value+gap;
			d = mat[j][i-1].value+gap;
			m = mat[j-1][i-1].value+1.0f-nu*sum;
			if (c<0.0f && m<0.0f && d<0.0f){
				mat[j][i].value = 0.0f;
				mat[j][i].move = 0;
			}else{
				if (c>m){
					if(c>d){
						mat[j][i].value = c;
						mat[j][i].move = 1;
					}else{
						mat[j][i].value = d;
						mat[j][i].move = 3;
					}
				}else{
					if(m>d){
						mat[j][i].value = m;
						mat[j][i].move = 2;
					}else{
						mat[j][i].value = d;
						mat[j][i].move = 3;
					}
				}
			}
			if(mat[j][i].value>result){
				result = mat[j][i].value;
				mi = i;
				mj = j;
			}
		}
	}
	string a1,a2,a3;
	int cptM = 0;
	int aliL = 1;
	vector<point> v1,v2;
	while(mat[mj][mi].value>0){
		if(mat[mj][mi].move==2){
			a1.insert(0,1,s1.residues[mi]);
			a2.insert(0,1,s2.residues[mj]);
			v1.push_back(s1.coords[mi]);
			v2.push_back(s2.coords[mj]);
			aliL += 1;
			if (s1.residues[mi]==s2.residues[mj]){
				cptM += 1;
				a3.insert(0,1,s2.residues[mj]);
			}else
				a3.insert(0,1,'.');
			mj-=1;
			mi-=1;
		}else if(mat[mj][mi].move==1){
			a1.insert(0,1,'-');
			a2.insert(0,1,s2.residues[mj]);
			a3.insert(0,1,'.');
			mj-=1;
		}else if(mat[mj][mi].move==3){
			a1.insert(0,1,s1.residues[mi]);
			a2.insert(0,1,'-');
			a3.insert(0,1,'.');
			mi-=1;			
		}
	}
	a1.insert(0,1,s1.residues[mi]);
	a2.insert(0,1,s2.residues[mj]);
	v1.push_back(s1.coords[mi]);
	v2.push_back(s2.coords[mj]);
	if (s1.residues[mi]==s2.residues[mj]){
		cptM += 1;
		a3.insert(0,1,s2.residues[mj]);
	}else
		a3.insert(0,1,' ');
	string l1 = convertInt(mi);
	string l2 = convertInt(mj);
	int len = max(l1.length(),l2.length());
	a1.insert(0,1+len-l1.length(),' ');
	a2.insert(0,1+len-l2.length(),' ');
	a3.insert(0,1+len,' ');
	a1.insert(0,convertInt(mi));
	a2.insert(0,convertInt(mj));
	for (int i=0;i<s2.size;i++)
		delete mat[i];
	delete mat;
	cout << a1 << endl << a2 << endl << a3 <<endl;
	cout << "Alignment length: "<< aliL << endl;
	cout << "Percentage of matching residues: "<< cptM*100.0f/aliL << " %"<< endl; 
	cout << "Score: "<< result<< endl;

	double p1[v1.size()][3], p2[v1.size()][3];
	for (int i=0;i<v1.size();i++){
		p1[i][0] = v1[i].x;
		p1[i][1] = v1[i].y;
		p1[i][2] = v1[i].z;
		p2[i][0] = v2[i].x;
		p2[i][1] = v2[i].y;
		p2[i][2] = v2[i].z;
	}
	double rmsd;
	fast_rmsd(p1,p2,v1.size(),&rmsd);
	cout << "RMSD: "<< rmsd<< endl;
}

void divideByAvg(unsigned int dimension,float *seq,unsigned int sOffset, unsigned int sLength){
	for (unsigned int k=0;k<dimension;k++){
		float avg = 0.0f;
		for (unsigned int i=0;i<sLength;i++)
			avg += seq[sOffset+k+i*dimension];
		avg /= sLength;
		for (unsigned int i=0;i<sLength;i++)
			seq[sOffset+k+i*dimension] = seq[sOffset+k+i*dimension]/avg;
	}
}

seq loadPDB(char *fileName,int nSigmas,float* sigmas){
	ifstream file(fileName);
	string line;
	vector<point> residuePos;
	string currNum = "";
	seq ret;
	float *dists,*totWei;
	float lx,ly,lz;
	if (file){
		while ( getline( file, line ) ){
			if (line.substr(0,3) == "TER")
				break;
			else if (line.length()>55 && line.substr(13,2) == "CA" && (line.substr(0,4) == "ATOM" || line.substr(0,6) == "HETATM") &&  line.substr(22,4) != currNum ){
				point p;
				p.x = atof(line.substr(30,8).c_str());
				p.y = atof(line.substr(38,8).c_str());
				p.z = atof(line.substr(46,8).c_str());
				residuePos.push_back(p);
				ret.residues.append(1,letterOfResidue(line.substr(17,3)));
				currNum = line.substr(22,4);
			}
		}
		ret.values = (float *)malloc(sizeof(float)*residuePos.size()*nSigmas);
		ret.coords = (point *)malloc(sizeof(point)*residuePos.size());
		dists = (float *)malloc(sizeof(float)*residuePos.size());
		totWei = (float *)malloc(sizeof(float)*nSigmas);
		for (int i=0;i<residuePos.size();i++){
			ret.coords[i] = residuePos[i];
			for (int j=0;j<nSigmas;j++)
				totWei[j] = 0.0f;
			for (int j=0;j<residuePos.size();j++)
				if(abs(i-j)>1){
					dists[j] = sqrt((residuePos[i].x-residuePos[j].x)*(residuePos[i].x-residuePos[j].x)+(residuePos[i].y-residuePos[j].y)*(residuePos[i].y-residuePos[j].y)+(residuePos[i].z-residuePos[j].z)*(residuePos[i].z-residuePos[j].z));
					for (int k=0;k<nSigmas;k++)
						totWei[k] += EXP(-(dists[j]*dists[j])/(sigmas[k]*sigmas[k]));

				}
			for (int j=0;j<nSigmas;j++){
				if (totWei[j]>0){
					lx = 0.0f;
					ly = 0.0f;
					lz = 0.0f;
					for (int k=0;k<residuePos.size();k++)
						if(abs(i-k)>1){
							lx += residuePos[k].x*EXP(-(dists[k]*dists[k])/(sigmas[j]*sigmas[j]));
							ly += residuePos[k].y*EXP(-(dists[k]*dists[k])/(sigmas[j]*sigmas[j]));
							lz += residuePos[k].z*EXP(-(dists[k]*dists[k])/(sigmas[j]*sigmas[j]));
						}
					lx = residuePos[i].x - lx/totWei[j];
					ly = residuePos[i].y - ly/totWei[j];
					lz = residuePos[i].z - lz/totWei[j];
					ret.values[i*nSigmas+j] = sqrt(lx*lx + ly*ly + lz*lz);
				}else
					ret.values[i*nSigmas+j] = 0.0f;
			}
		}
		free(dists);
		free(totWei);
		ret.size = residuePos.size();
		return ret;
	}else
		cout << "Error, PDB file not found"<<endl; 	
}

int main(int argc, char *argv[]){
	if (argc != 4 or (strcmp(argv[1], "-nw1")!=0 and strcmp(argv[1], "-sw1")!=0 and strcmp(argv[1], "-nw2")!=0 and strcmp(argv[1], "-sw2")!=0)){
		cout << "Usage is \"lnaOvO [-method] [PDB file name] [PDB file name] with method = nw or sw"<<endl;
		exit(1);
	}
	seq p1,p2;
	if (strcmp(argv[1], "-nw1")==0){
		float sigmas[] = {6.1f};
		p1 = loadPDB(argv[3],1,sigmas);
		p2 = loadPDB(argv[2],1,sigmas);
		nwa(1,p1,p2,0.24f);
	}else if (strcmp(argv[1], "-nw2")==0){
		float sigmas[] = {5.4f,14.3f};
		p1 = loadPDB(argv[3],2,sigmas);
		p2 = loadPDB(argv[2],2,sigmas);
		nwa(2,p1,p2,0.15f);
	}else if (strcmp(argv[1], "-sw1")==0){
		float sigmas[] = {5.7f};
		p1 = loadPDB(argv[3],1,sigmas);
		p2 = loadPDB(argv[2],1,sigmas);
		divideByAvg(1,p1.values,0,p1.size);
		divideByAvg(1,p2.values,0,p2.size);
		swa(1,p1,p2,0.67f,-0.53f);
	}else{
		float sigmas[] = {5.0f,14.5f};
		p1 = loadPDB(argv[3],2,sigmas);
		p2 = loadPDB(argv[2],2,sigmas);
		divideByAvg(2,p1.values,0,p1.size);
		divideByAvg(2,p2.values,0,p2.size);
		swa(2,p1,p2,0.41f,-0.5f);
	}
	free(p1.values);
	free(p2.values);
	free(p1.coords);
	free(p2.coords);
	return 0;
}

